@@ -37,6 +37,7 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         self.drac_client = dracclient.client.DRACClient(
             **test_utils.FAKE_ENDPOINT)
         self.raid_controller_fqdd = "RAID.Integrated.1-1"
+        self.boss_controller_fqdd = "AHCI.Slot.3-1"
         cntl_dict = {'RAID.Integrated.1-1':
                      ['Disk.Bay.0:Enclosure.Internal.0-1:RAID.Integrated.1-1',
                       'Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1'],
@@ -586,6 +587,131 @@ class ClientRAIDManagementTestCase(base.BaseTest):
             exceptions.DRACOperationFailed,
             self.drac_client.delete_virtual_disk, 'disk1')
 
+    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
+                       spec_set=True, autospec=True)
+    def test_reset_raid_config(self, mock_requests, mock_invoke):
+        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
+                              'CreationClassName': 'DCIM_RAIDService',
+                              'SystemName': 'DCIM:ComputerSystem',
+                              'Name': 'DCIM:RAIDService'}
+        expected_properties = {'Target': self.raid_controller_fqdd}
+        mock_invoke.return_value = lxml.etree.fromstring(
+            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
+                'ResetConfig']['ok'])
+        result = self.drac_client.reset_raid_config(self.raid_controller_fqdd)
+        self.assertEqual({'is_commit_required': True,
+                          'is_reboot_required':
+                          constants.RebootRequired.optional},
+                         result)
+        mock_invoke.assert_called_once_with(
+            mock.ANY, uris.DCIM_RAIDService, 'ResetConfig',
+            expected_selectors, expected_properties,
+            expected_return_value=utils.RET_SUCCESS)
+
+    @mock.patch.object(dracclient.client.WSManClient,
+                       'wait_until_idrac_is_ready', spec_set=True,
+                       autospec=True)
+    def test_reset_raid_config_fail(self, mock_requests,
+                                    mock_wait_until_idrac_is_ready):
+        mock_requests.post(
+            'https://1.2.3.4:443/wsman',
+            text=test_utils.RAIDInvocations[
+                uris.DCIM_RAIDService]['ResetConfig']['error'])
+
+        self.assertRaises(
+            exceptions.DRACOperationFailed,
+            self.drac_client.reset_raid_config, self.raid_controller_fqdd)
+
+    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
+                       spec_set=True, autospec=True)
+    def test_clear_foreign_config(self, mock_requests, mock_invoke):
+        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
+                              'CreationClassName': 'DCIM_RAIDService',
+                              'SystemName': 'DCIM:ComputerSystem',
+                              'Name': 'DCIM:RAIDService'}
+        expected_properties = {'Target': self.raid_controller_fqdd}
+        mock_invoke.return_value = lxml.etree.fromstring(
+            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
+                'ClearForeignConfig']['ok'])
+
+        result = self.drac_client.clear_foreign_config(
+                self.raid_controller_fqdd)
+        self.assertEqual({'is_commit_required': True,
+                          'is_reboot_required':
+                          constants.RebootRequired.optional},
+                         result)
+        mock_invoke.assert_called_once_with(
+            mock.ANY, uris.DCIM_RAIDService, 'ClearForeignConfig',
+            expected_selectors, expected_properties,
+            check_return_value=False)
+
+    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
+                       spec_set=True, autospec=True)
+    def test_clear_foreign_config_with_no_foreign_drive(self,
+                                                        mock_requests,
+                                                        mock_invoke):
+        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
+                              'CreationClassName': 'DCIM_RAIDService',
+                              'SystemName': 'DCIM:ComputerSystem',
+                              'Name': 'DCIM:RAIDService'}
+        expected_properties = {'Target': self.raid_controller_fqdd}
+        mock_invoke.return_value = lxml.etree.fromstring(
+            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
+                'ClearForeignConfig']['no_foreign_drive'])
+
+        result = self.drac_client.clear_foreign_config(
+                self.raid_controller_fqdd)
+        self.assertEqual({'is_commit_required': False,
+                          'is_reboot_required':
+                          constants.RebootRequired.false},
+                         result)
+        mock_invoke.assert_called_once_with(
+            mock.ANY, uris.DCIM_RAIDService, 'ClearForeignConfig',
+            expected_selectors, expected_properties,
+            check_return_value=False)
+
+    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
+                       spec_set=True, autospec=True)
+    def test_clear_foreign_config_with_operation_not_supported(self,
+                                                               mock_requests,
+                                                               mock_invoke):
+        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
+                              'CreationClassName': 'DCIM_RAIDService',
+                              'SystemName': 'DCIM:ComputerSystem',
+                              'Name': 'DCIM:RAIDService'}
+        expected_properties = {'Target': self.boss_controller_fqdd}
+        mock_invoke.return_value = lxml.etree.fromstring(
+            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
+                'ClearForeignConfig']['foreign_drive_operation_not_supported'])
+
+        result = self.drac_client.clear_foreign_config(
+                self.boss_controller_fqdd)
+        self.assertEqual({'is_commit_required': False,
+                          'is_reboot_required':
+                          constants.RebootRequired.false},
+                         result)
+        mock_invoke.assert_called_once_with(
+            mock.ANY, uris.DCIM_RAIDService, 'ClearForeignConfig',
+            expected_selectors, expected_properties,
+            check_return_value=False)
+
+    @mock.patch.object(dracclient.client.WSManClient,
+                       'wait_until_idrac_is_ready', spec_set=True,
+                       autospec=True)
+    def test_clear_foreign_config_with_invalid_controller_id(
+            self,
+            mock_requests,
+            mock_wait_until_idrac_is_ready):
+        mock_requests.post(
+            'https://1.2.3.4:443/wsman',
+            text=test_utils.RAIDInvocations[
+                uris.DCIM_RAIDService]['ClearForeignConfig']
+            ['invalid_controller_id'])
+
+        self.assertRaises(
+            exceptions.DRACOperationFailed,
+            self.drac_client.clear_foreign_config, 'bad')
+
     @mock.patch.object(dracclient.resources.job.JobManagement,
                        'create_config_job', spec_set=True, autospec=True)
     def test_commit_pending_raid_changes(self, mock_requests,
@@ -799,9 +925,29 @@ class ClientRAIDManagementTestCase(base.BaseTest):
             exceptions.DRACOperationFailed,
             self.drac_client.is_jbod_capable, self.raid_controller_fqdd)
 
-    def test_is_raid_controller(self, mock_requests):
+    def test_is_raid_controller_raid(self, mock_requests):
         self.assertTrue(self.drac_client
                         .is_raid_controller("RAID.Integrated.1-1"))
+
+    @mock.patch.object(dracclient.client.WSManClient,
+                       'wait_until_idrac_is_ready', spec_set=True,
+                       autospec=True)
+    def test_is_raid_controller_boss(self, mock_requests,
+                                     mock_wait_until_idrac_is_ready):
+        mock_requests.post(
+            'https://1.2.3.4:443/wsman',
+            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])
+        self.assertTrue(self.drac_client
+                        .is_raid_controller("AHCI.Integrated.1-1"))
+
+    @mock.patch.object(dracclient.client.WSManClient,
+                       'wait_until_idrac_is_ready', spec_set=True,
+                       autospec=True)
+    def test_is_raid_controller_fail(self, mock_requests,
+                                     mock_wait_until_idrac_is_ready):
+        mock_requests.post(
+            'https://1.2.3.4:443/wsman',
+            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])
         self.assertFalse(self.drac_client
                          .is_raid_controller("notRAID.Integrated.1-1"))
 
@@ -871,13 +1017,13 @@ class ClientRAIDManagementTestCase(base.BaseTest):
                           raid_mgt._check_disks_status,
                           mode,
                           physical_disks,
-                          self.controllers_to_physical_disk_ids.copy())
+                          self.controllers_to_physical_disk_ids)
         mode = constants.RaidStatus.jbod
         self.assertRaises(ValueError,
                           raid_mgt._check_disks_status,
                           mode,
                           physical_disks,
-                          self.controllers_to_physical_disk_ids.copy())
+                          self.controllers_to_physical_disk_ids)
 
     def test_check_disks_status_fail(self, mock_requests):
         mode = constants.RaidStatus.raid
@@ -889,13 +1035,13 @@ class ClientRAIDManagementTestCase(base.BaseTest):
                           raid_mgt._check_disks_status,
                           mode,
                           physical_disks,
-                          self.controllers_to_physical_disk_ids.copy())
+                          self.controllers_to_physical_disk_ids)
         mode = constants.RaidStatus.jbod
         self.assertRaises(ValueError,
                           raid_mgt._check_disks_status,
                           mode,
                           physical_disks,
-                          self.controllers_to_physical_disk_ids.copy())
+                          self.controllers_to_physical_disk_ids)
 
     def test_check_disks_status_no_change(self, mock_requests):
         raid_mgt = self.drac_client._raid_mgmt
@@ -903,11 +1049,8 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         physical_disks = [self.disk_1, self.disk_2,
                           self.disk_3, self.disk_4]
 
-        raid_cntl_to_phys_disk_ids = (self.controllers_to_physical_disk_ids.
-                                      copy())
-
-        raid_mgt._check_disks_status(mode,  physical_disks,
-                                     raid_cntl_to_phys_disk_ids)
+        raid_cntl_to_phys_disk_ids = raid_mgt._check_disks_status(
+            mode,  physical_disks, self.controllers_to_physical_disk_ids)
         raid_len = len(raid_cntl_to_phys_disk_ids['RAID.Integrated.1-1'])
         self.assertEqual(raid_len, 0)
 
@@ -917,10 +1060,8 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         physical_disks = [disk_1_non_raid, disk_2_non_raid,
                           self.disk_3, self.disk_4]
 
-        jbod_cntl_to_phys_disk_ids = (self.controllers_to_physical_disk_ids.
-                                      copy())
-        raid_mgt._check_disks_status(mode,  physical_disks,
-                                     jbod_cntl_to_phys_disk_ids)
+        jbod_cntl_to_phys_disk_ids = raid_mgt._check_disks_status(
+            mode,  physical_disks, self.controllers_to_physical_disk_ids)
         jbod_len = len(jbod_cntl_to_phys_disk_ids['RAID.Integrated.1-1'])
         self.assertEqual(jbod_len, 0)
 
@@ -929,11 +1070,8 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         mode = constants.RaidStatus.jbod
         physical_disks = [self.disk_1, self.disk_2, self.disk_3, self.disk_4]
 
-        jbod_cntl_to_phys_disk_ids = (self.controllers_to_physical_disk_ids.
-                                      copy())
-
-        raid_mgt._check_disks_status(mode, physical_disks,
-                                     jbod_cntl_to_phys_disk_ids)
+        jbod_cntl_to_phys_disk_ids = raid_mgt._check_disks_status(
+            mode, physical_disks, self.controllers_to_physical_disk_ids)
         jbod_len = len(jbod_cntl_to_phys_disk_ids['RAID.Integrated.1-1'])
         self.assertEqual(jbod_len, 2)
 
@@ -942,10 +1080,8 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         disk_2_non_raid = self.disk_2._replace(raid_status='non-RAID')
         physical_disks = [disk_1_non_raid, disk_2_non_raid,
                           self.disk_3, self.disk_4]
-        raid_cntl_to_phys_disk_ids = (self.controllers_to_physical_disk_ids.
-                                      copy())
-        raid_mgt._check_disks_status(mode,  physical_disks,
-                                     raid_cntl_to_phys_disk_ids)
+        raid_cntl_to_phys_disk_ids = raid_mgt._check_disks_status(
+            mode,  physical_disks, self.controllers_to_physical_disk_ids)
         raid_len = len(raid_cntl_to_phys_disk_ids['RAID.Integrated.1-1'])
         self.assertEqual(raid_len, 2)
 
@@ -960,13 +1096,13 @@ class ClientRAIDManagementTestCase(base.BaseTest):
                           raid_mgt._check_disks_status,
                           mode,
                           physical_disks,
-                          self.controllers_to_physical_disk_ids.copy())
+                          self.controllers_to_physical_disk_ids)
         mode = constants.RaidStatus.jbod
         self.assertRaises(ValueError,
                           raid_mgt._check_disks_status,
                           mode,
                           physical_disks,
-                          self.controllers_to_physical_disk_ids.copy())
+                          self.controllers_to_physical_disk_ids)
 
     @mock.patch.object(dracclient.client.WSManClient,
                        'wait_until_idrac_is_ready', spec_set=True,
@@ -982,15 +1118,21 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         mock_requests.post(
             'https://1.2.3.4:443/wsman',
             text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])
-        mock_convert_physical_disks.return_value = {'is_commit_required': True,
-                                                    'is_reboot_required':
-                                                    constants.RebootRequired
-                                                    .true}
-        cntl_to_phys_d_ids = self.controllers_to_physical_disk_ids
+        cvt_phys_disks_return_value = {'is_commit_required': True,
+                                       'is_reboot_required': constants.
+                                       RebootRequired.true}
+        mock_convert_physical_disks.return_value = cvt_phys_disks_return_value
+
+        expected_return_value = {'RAID.Integrated.1-1':
+                                 cvt_phys_disks_return_value,
+                                 'AHCI.Integrated.1-1':
+                                 cvt_phys_disks_return_value}
         results = self.drac_client.change_physical_disk_state(
-            mode, cntl_to_phys_d_ids)
+            mode, self.controllers_to_physical_disk_ids)
         self.assertTrue(results["is_reboot_required"])
         self.assertEqual(len(results["commit_required_ids"]), 2)
+        self.assertEqual(results['conversion_results'],
+                         expected_return_value)
 
     @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                        'list_physical_disks', spec_set=True,
@@ -1008,38 +1150,44 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         physical_disks = [disk_1_non_raid, disk_2_non_raid,
                           self.disk_3, self.disk_4]
         mock_list_physical_disks.return_value = physical_disks
-        mock_convert_physical_disks.return_value = {'is_commit_required': True,
-                                                    'is_reboot_required':
-                                                    constants.RebootRequired
-                                                    .true}
-        cntl_to_phys_d_ids = self.controllers_to_physical_disk_ids
+        boss_return_value = {'is_commit_required': False,
+                             'is_reboot_required':
+                             constants.RebootRequired.false}
+        raid_return_value = {'is_commit_required': True,
+                             'is_reboot_required':
+                             constants.RebootRequired.true}
+        mock_convert_physical_disks.return_value = raid_return_value
+
         results = self.drac_client.change_physical_disk_state(
-            mode, cntl_to_phys_d_ids)
+            mode, self.controllers_to_physical_disk_ids)
         self.assertTrue(results["is_reboot_required"])
         self.assertEqual(len(results["commit_required_ids"]), 1)
+        self.assertEqual(len(results['conversion_results']), 2)
+        self.assertEqual(results['conversion_results']['AHCI.Integrated.1-1'],
+                         boss_return_value)
+        self.assertEqual(results['conversion_results']['RAID.Integrated.1-1'],
+                         raid_return_value)
 
     @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                        'list_physical_disks', spec_set=True,
                        autospec=True)
-    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
-                       'convert_physical_disks', spec_set=True,
-                       autospec=True)
     def test_change_physical_disk_state_none(
             self, mock_requests,
-            mock_convert_physical_disks,
             mock_list_physical_disks):
         mode = constants.RaidStatus.raid
         physical_disks = [self.disk_1, self.disk_2, self.disk_3, self.disk_4]
-        mock_convert_physical_disks.return_value = {'is_commit_required': True,
-                                                    'is_reboot_required':
-                                                    constants.RebootRequired
-                                                    .true}
         mock_list_physical_disks.return_value = physical_disks
-        cntl_to_phys_d_ids = self.controllers_to_physical_disk_ids
+        expected_return_value = {'is_commit_required': False,
+                                 'is_reboot_required':
+                                 constants.RebootRequired.false}
         results = self.drac_client.change_physical_disk_state(
-            mode, cntl_to_phys_d_ids)
+            mode, self.controllers_to_physical_disk_ids)
         self.assertFalse(results["is_reboot_required"])
         self.assertEqual(len(results["commit_required_ids"]), 0)
+        self.assertEqual(results['conversion_results']['RAID.Integrated.1-1'],
+                         expected_return_value)
+        self.assertEqual(results['conversion_results']['AHCI.Integrated.1-1'],
+                         expected_return_value)
 
     @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                        'list_physical_disks', spec_set=True,
@@ -1059,11 +1207,17 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         physical_disks = [disk_1_non_raid, disk_2_non_raid,
                           self.disk_3, self.disk_4]
         mock_list_physical_disks.return_value = physical_disks
-        cntl_to_phys_d_ids = self.controllers_to_physical_disk_ids
+        expected_return_value = {'is_commit_required': False,
+                                 'is_reboot_required':
+                                 constants.RebootRequired.false}
         results = self.drac_client.change_physical_disk_state(
-            mode, cntl_to_phys_d_ids)
+            mode, self.controllers_to_physical_disk_ids)
         self.assertFalse(results["is_reboot_required"])
         self.assertEqual(len(results["commit_required_ids"]), 0)
+        self.assertEqual(results['conversion_results']['RAID.Integrated.1-1'],
+                         expected_return_value)
+        self.assertEqual(results['conversion_results']['AHCI.Integrated.1-1'],
+                         expected_return_value)
 
     @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                        'list_physical_disks', spec_set=True,
@@ -1083,12 +1237,12 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         physical_disks = [disk_1_non_raid, disk_2_non_raid,
                           self.disk_3, self.disk_4]
         mock_list_physical_disks.return_value = physical_disks
-        cntl_to_phys_d_ids = self.controllers_to_physical_disk_ids
         self.assertRaisesRegexp(
             exceptions.DRACOperationFailed,
             "OTHER_MESSAGE",
             self.drac_client.change_physical_disk_state,
-            mode, cntl_to_phys_d_ids)
+            mode,
+            self.controllers_to_physical_disk_ids)
 
     @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                        'list_physical_disks', spec_set=True,
@@ -1107,11 +1261,12 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         physical_disks = [disk_1_non_raid, disk_2_non_raid,
                           self.disk_3, self.disk_4]
         mock_list_physical_disks.return_value = physical_disks
-        cntl_to_phys_d_ids = self.controllers_to_physical_disk_ids
         self.assertRaisesRegexp(
-            Exception, "SOMETHING_BAD_HAPPENED",
+            Exception,
+            "SOMETHING_BAD_HAPPENED",
             self.drac_client.change_physical_disk_state,
-            mode, cntl_to_phys_d_ids)
+            mode,
+            self.controllers_to_physical_disk_ids)
 
     @mock.patch.object(dracclient.client.WSManClient,
                        'wait_until_idrac_is_ready', spec_set=True,
@@ -1165,45 +1320,6 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         self.assertFalse(results["is_reboot_required"])
         self.assertEqual(len(results["commit_required_ids"]), 0)
 
-    @mock.patch.object(dracclient.client.WSManClient,
-                       'wait_until_idrac_is_ready', spec_set=True,
-                       autospec=True)
-    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
-                       'list_physical_disks', spec_set=True,
-                       autospec=True)
-    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
-                       'convert_physical_disks', spec_set=True,
-                       autospec=True)
-    def test_change_physical_disk_state_conversion_return_values(
-            self, mock_requests,
-            mock_convert_physical_disks,
-            mock_list_physical_disks,
-            mock_wait_until_idrac_is_ready):
-        mock_requests.post(
-            'https://1.2.3.4:443/wsman',
-            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])
-        mode = constants.RaidStatus.jbod
-        physical_disks = [self.disk_1, self.disk_2, self.disk_3, self.disk_4]
-        '''Test all logic branches for 100% coverage, it is unlikely
-        convert_physical_disks() will return empty dict but we do check
-        for this case in change_physical_disk_state()'''
-        mock_convert_physical_disks.return_value = {}
-        mock_list_physical_disks.return_value = physical_disks
-        results = self.drac_client.change_physical_disk_state(mode)
-        self.assertFalse(results["is_reboot_required"])
-        self.assertEqual(len(results["commit_required_ids"]), 0)
-        '''Where convert_physical_disks() does not require a commit after
-        executing, unlikely case but provides 100% code coverage of all
-        logic branches.'''
-        mock_convert_physical_disks.return_value = {'is_commit_required':
-                                                    False,
-                                                    'is_reboot_required':
-                                                    constants.RebootRequired
-                                                    .false}
-        results = self.drac_client.change_physical_disk_state(mode)
-        self.assertFalse(results["is_reboot_required"])
-        self.assertEqual(len(results["commit_required_ids"]), 0)
-
     @mock.patch.object(dracclient.client.WSManClient,
                        'wait_until_idrac_is_ready', spec_set=True,
                        autospec=True)
